var ZombieKeys = new function() {
    var debugLevel = 0;

    // modifiers: [ctrl, shift, alt]
    var deadKeys= [{"modifiers": {"ctrlKey": true, "shiftKey": false, 
				  "altKey": false},
		    "keyCode": 192, // "`"
		    "mapping": {"a": "\u00e0",
				"e": "\u00e8",
				"i": "\u00ec",
				"o": "\u00f2",
				"u": "\u00f9",
				"A": "\u00c0",
				"E": "\u00c8",
				"I": "\u00cc",
				"O": "\u00d2",
				"U": "\u00d9"}},
		 
	           {"modifiers": {"ctrlKey": true, "shiftKey": false, 
				  "altKey": false},
		    "keyCode": 222, // "'" 
		    "mapping": {"a": "\u00e1",
				"e": "\u00e9",
				"i": "\u00ed",
				"o": "\u00f3",
				"u": "\u00fa",
				"y": "\u00fd",
				"A": "\u00c1",
				"E": "\u00c9",
				"I": "\u00cd",
				"O": "\u00d3",
				"U": "\u00da",
				"Y": "\u00dd",
				"d": "\u00f0",
				"D": "\u00d0"}},  

                   {"modifiers": {"ctrlKey": true, "shiftKey":  true, 
				  "altKey": false},
		    "keyCode": 54, // "^"
		    "mapping": {"a": "\u00e2",
				"e": "\u00ea",
				"i": "\u00ee",
				"o": "\u00f4",
				"u": "\u00fb",
				"A": "\u00c2",
				"E": "\u00ca",
				"I": "\u00ce",
				"O": "\u00d4",
				"U": "\u00db"}},

                   {"modifiers": {"ctrlKey": true, "shiftKey":  true, 
				  "altKey": false},
		    "keyCode": 192, // "~"
		    "mapping": {"a": "\u00e3",
				"n": "\u00f1",
				"o": "\u00f5",
				"A": "\u00c3",
				"N": "\u00d1",
				"O": "\u00d5"}},
		 
                   {"modifiers": {"ctrlKey": true, "shiftKey":  true, 
				  "altKey": false},
		    "keyCode": 59, // ":" 
		    "mapping": {"a": "\u00e4",
				"e": "\u00eb",
				"i": "\u00ef",
				"o": "\u00f6",
				"u": "\u00fc",
				"y": "\u00ff",
				"A": "\u00c4",
				"E": "\u00cb",
				"I": "\u00cf",
				"O": "\u00d6",
				"U": "\u00dc",
				"Y": "\u0178"}},

	           {"modifiers": {"ctrlKey": true, "shiftKey":  true, 
				  "altKey": false},
		    "keyCode": 50, // "@"
		    "mapping": {"a": "\u00e5",
				"A": "\u00c5"}},

                   {"modifiers": {"ctrlKey": true, "shiftKey":  true, 
				  "altKey": false},
		    "keyCode": 55, // "&"
		    "mapping": {"a": "\u00e6",
				"A": "\u00c6",
				"o": "\u0153",
				"O": "\u0152",
				"s": "\u00df"}},

                   {"modifiers": {"ctrlKey": true, "shiftKey": false, 
				  "altKey": false},
		    "keyCode": 188, // ","
		    "mapping": {"c": "\u00e7",
				"C": "\u00c7"}},

                   {"modifiers": {"ctrlKey": true, "shiftKey": false, 
				  "altKey": false},
		    "keyCode": 191, // "/"
		    "mapping": {"o": "\u00f8",
				"O": "\u00d8",
				"c": "\u00a2"}}];

    var aliveKeys = [{"modifiers": {"ctrlKey": true, "shiftKey":  true, 
				    "altKey":  true}, 
		      "keyCode": 191, // "?"
		      "target": "\u00bf"},
	             {"modifiers": {"ctrlKey": true, "shiftKey":  true, 
				    "altKey":  true}, 
		      "keyCode": 49, // "!"
		      "target": "\u00a1"},
                     {"modifiers": {"ctrlKey": true, "shiftKey": false, 
				    "altKey":  true}, 
		      "charCode": 101, // "e"
		      "target": "\u20ac"}];

    var hexUnicodeKey = {"modifiers": {"ctrlKey": false, "shiftKey": false, 
				       "altKey":  true}, 
			 "keyCode": 88}; // "x"

    var decUnicodeModifiers = {"ctrlKey": false, "shiftKey": false, 
			       "altKey":  true};

    var decUnicode = "";

    var keyEventProps = ['keyCode', 'charCode', 'shiftKey', 
			 'altKey', 'ctrlKey'];

    function log(message) {
	var consoleService = Components.classes["@mozilla.org/consoleservice;1"]
	    .getService(Components.interfaces.nsIConsoleService);
	consoleService.logStringMessage(message + (message != "\n" ? " " : ""));
    };

    function logKey(why, event) {
	var message=why;
	for (var k=0; k<keyEventProps.length; k++) {
	    message += keyEventProps[k] + "=" + event[keyEventProps[k]]+" ";
	}
	log(message+"\n");
    }

    function logMapping(mapping) {
	var message="";
	for (var mappee in mapping) {
	    message += mappee+" ("+mappee.charCodeAt(0)+") => "+
		mapping[mappee]+" ("+mapping[mappee].charCodeAt(0)+") ";
	}        
	log(message+"\n");
    }

    function checkModifiers(event, modifiers) {
	for (var key in modifiers) {
	    if (event[key] != modifiers[key]) {
		return false;
	    }
	}
	return true;
    }

    function keyPressHandler(event) {
	if (debugLevel > 5) {logKey("press: ", event);}

	// check "alive" keys which fire during "keypress"
	for (var k=0; k<aliveKeys.length; k++) {
	    if (aliveKeys[k].charCode && 
		(event.charCode == aliveKeys[k].charCode) &&
		checkModifiers(event, aliveKeys[k].modifiers)) {
		fakeKey(event, aliveKeys[k].target);
		return;
	    }
	}

	// check if/which "dead" key is active
	for (var k=0; k<deadKeys.length; k++) {
	    if (deadKeys[k].alive) {
		deadKeys[k].alive = false;
		// check if there's a mapping for this character 
		// entered after "dead" key
		if (debugLevel > 10) {
		    logMapping("detected zombie #"+k+" for "+event.charCode+
			       " => "+deadKeys[k].mapping);
		}
		var code = String.fromCharCode(event.charCode);
		if (deadKeys[k].mapping[code]) {
		    fakeKey(event, deadKeys[k].mapping[code]);
		    return;
		}
	    }
	}
    };

    function keyUpHandler(event) {
	if (debugLevel > 5) {logKey("up: ", event);}

	// check "alive" keys which only fire on "keyup"
	for (var k=0; k<aliveKeys.length; k++) {
	    if (aliveKeys[k].keyCode && 
		(event.keyCode == aliveKeys[k].keyCode) &&
		checkModifiers(event, aliveKeys[k].modifiers)) {
		fakeKey(event, aliveKeys[k].target);
		return;
	    }
	}
	
	// check if "dead" key is activated
	for (var k=0; k<deadKeys.length; k++) {
	    if ((event.keyCode == deadKeys[k].keyCode) &&
		checkModifiers(event, deadKeys[k].modifiers)) {
		if (debugLevel > 10) {
		    logMapping("created zombie #"+k+": "+deadKeys[k].mapping);
		}
		deadKeys[k].alive = true;
		event.preventDefault();
		event.stopPropagation();
		return;
	    }
	}
	
	// process ALT+X (treat preceeding 4 chars as hex unicode string 
	// and replace with corresponding char)
	if ((event.keyCode == hexUnicodeKey.keyCode) &&
	    checkModifiers(event, hexUnicodeKey.modifiers)) {
	    var hexUnicode = getPreceedingString(event.target, 4);
	    var code = parseInt(hexUnicode, 16);
	    replacePreceedingString(event.target, 4, String.fromCharCode(code));
	    return;
	}
	
	// process ALT+<numeric keypad>
	if (checkModifiers(event, decUnicodeModifiers)) {
	    if ((event.keyCode >= 96) && (event.keyCode <= 105)) {
		decUnicode += (event.keyCode-96).toString(); // add digit
		event.preventDefault();
		event.stopPropagation();
		return;		
	    } else if (event.keyCode == 18) {
		// released ALT - let's insert
		if (decUnicode != "") {
		    var code = parseInt(decUnicode, 10);
		    fakeKey(event, String.fromCharCode(code));
		    decUnicode = "";
		}
		event.preventDefault();
		event.stopPropagation();
		return;
	    }
	}
    };

    function fakeKey(event, chr) {
	var new_event = document.createEvent("KeyEvents");
	// initKeyEvent(type, canBubble, cancelable, view, 
	//              ctrlKey, altKey, shiftKey, metaKey, keyCode, charCode);
	new_event.initKeyEvent("keypress", true, true, event.view, 
			       false, false, false, false, 
			       0, chr.charCodeAt(0));
	event.target.dispatchEvent(new_event); 
	if (debugLevel > 0) {logKey("faking '"+chr+"': ", new_event);}
	event.preventDefault();
	event.stopPropagation();
    }

    // get string preceeding cursor
    function getPreceedingString(elt, numchars) {
	var pos = elt.selectionStart;
	if (pos < numchars) {numchars=pos;}
	return elt.value.slice(pos-numchars,pos);
    }

    // replace string preceeding cursor
    function replacePreceedingString(elt, numchars, newtext) {
	var posStart = elt.selectionStart;   
	var posEnd = elt.selectionEnd;
	var scrollTop = elt.scrollTop;
	var scrollLeft = elt.scrollLeft;
	
	if (posStart < numchars) {numchars=posStart;}
	
	elt.value = elt.value.slice(0,posStart-numchars)+
                    newtext+
	            elt.value.slice(posEnd);
	var newpos = posStart+newtext.length;
	elt.selectionStart = newpos;
	elt.selectionEnd = newpos;
	
	elt.scrollTop = scrollTop;
	elt.scrollLeft = scrollLeft;
	elt.focus();
    }

    this.init = function(event) {
	if (window) {
	    window.addEventListener('keypress', keyPressHandler, false);
	    window.addEventListener('keyup',    keyUpHandler,    false);
	}
    }
};

if (window) {
    window.addEventListener("load", ZombieKeys.init, false);
}
